""" CCOBRA data container.

"""

import copy
import logging

from . import convert_to_basic_types
from .item import Item

# Initialize module-level logger
logger = logging.getLogger(__name__)

class CCobraData():
    """ CCobra experimental data container.

    """

    def __init__(self, data, target_columns):
        """ Initializes the CCOBRA data container by passing a data frame
        and validating its contents.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame to store in the CCOBRA data container.

        required_fields : list(str), optional
            List of required columns in the data. Defaults to ['id', 'sequence',
            'task', 'choices', 'response', 'response_type', 'domain']

        """

        self.target_columns = target_columns
        self.required_fields = [
            'id', 'sequence', 'task', 'choices', 'response_type', 'domain'
        ] + target_columns

        # Verify and store the data
        self.verify_data(data)
        self._data = data

        # Normalize the data container
        self.prepare_data()

        # Extract meta information
        self.n_subjects = len(self._data['_unique_id'].unique())
        self.domains = self._data['domain'].unique().tolist()
        self.response_types = self._data['response_type'].unique().tolist()

    def verify_data(self, data):
        """ Verifies if all required fields are in the data.

        Parameters
        ----------
        data : pd.DataFrame
            DataFrame to verify.

        Raises
        ------
        ValueError
            Thrown if data does not contain required columns.

        """

        missing = set(self.required_fields) - set(data.columns)
        if missing:
            raise ValueError(
                "Data does not contain columns: {}".format(missing))

    def prepare_data(self):
        """ Prepares the dataset by adding internally_used columns

        """

        assert '_unique_id' not in self._data

        # Add unique numerical subject identifier
        self._data['_unique_id'] = self._data['id']

    def prefix_identifiers(self, prefix='_train_'):
        """ Prefixes the subject identifier keys.

        Parameters
        ----------
        prefix : str
            Prefix to apply to key numerical identifiers.

        """

        self._data['_unique_id'] = self._data['_unique_id'].apply(lambda x: prefix + str(x))

    def get(self):
        """ Returns the contained data.

        Returns
        -------
        pd.DataFrame
            Dataframe containing the data.

        """

        return self._data

    def head(self):
        """ Displays the first 10 lines of the dataframe.

        """

        return self._data.head()

    def to_eval_dict(self):
        """ Converts the dataset to an evaluation dictionary mapping from individuals to data.

        Returns
        -------
        dict(object, list)
            Dictionary mapping from subject identifiers to lists of experimental data.

        """

        # Prepare the dictionary of subjects containing lists of tasks they responded to
        df = self._data

        dataset = {}
        for subj, subj_df in df.groupby('_unique_id'):
            assert subj not in dataset

            subj_df = subj_df.sort_values('sequence')

            subj_data = []
            for _, task_series in subj_df.iterrows():
                task_dict = {}

                # Extract the task information
                item = Item(
                    task_series['id'], task_series['domain'],
                    task_series['task'], task_series['response_type'],
                    task_series['choices'], task_series['sequence']
                )
                task_dict['item'] = item

                # Parse the main response
                responses = None
                if isinstance(task_series['response'], str):
                    responses = []
                    for response in task_series['response'].split('|'):
                        responses.append([x.split(';') for x in response.split('/')])
                    if task_series['response_type'] != 'multiple-choice':
                        responses = responses[0]
                else:
                    responses = task_series['response']
                task_dict['response'] = convert_to_basic_types(responses)

                # Parse the auxiliary targets
                for target_col in self.target_columns:
                    if target_col == 'response':
                        continue

                    if isinstance(task_series[target_col], str):
                        responses = []
                        for response in task_series[target_col].split('|'):
                            responses.append([x.split(';') for x in response.split('/')])
                    else:
                        responses = task_series[target_col]
                    task_dict[target_col] = responses

                # Add auxiliary elements from the data
                aux = {}
                for key, value in task_series.iteritems():
                    if key not in self.required_fields + ['_unique_id']:
                        aux[key] = value
                task_dict['aux'] = aux

                task_dict['full'] = copy.deepcopy(task_dict['aux'])
                for target_col in self.target_columns:
                    task_dict['full'][target_col] = task_dict[target_col]

                subj_data.append(task_dict)
            dataset[subj] = subj_data

        return dataset



""" No Valid Conclusion (NVC) comparator.

"""

from ccobra import CCobraComparator, tuple_to_string


class NVCComparator(CCobraComparator):
    """ NVC response comparator. Performs the evaluation based on NVC and non-NVC classes.

    """

    def compare(self, prediction, target, response_type, choices):
        """ Compares two response objects based on their NVCness. Only returns true if both
        responses are in agreement with either responding NVC or not NVC.

        Parameters
        ----------
        prediction : tuple
            Response tuple A for comparison.

        target : tuple
            Response tuple B for comparison.
            
        response_type : string
            The response type of the prediction and target.
            
        choices : list(object)
            The choice options that were available for this comparison.

        Returns
        -------
        bool
            True only if both objects agree on whether the response is NVC or not.

        """

        if response_type == "multiple-choice":
            raise ValueError('NVC Accuracy Comparator is incompatible with the multiple-choice response type.')

        is_nvc_a = tuple_to_string(prediction) == 'NVC'
        is_nvc_b = tuple_to_string(target) == 'NVC'
        return int(is_nvc_a == is_nvc_b)

    def get_name(self):
        """ Returns the name of the comparator.

        Returns
        -------
        string
            Comparator name.

        """

        return 'NVC Accuracy'



""" Equality comparator.

"""

import numpy as np

from ccobra import CCobraComparator, tuple_to_string


class EqualityComparator(CCobraComparator):
    """ Equality comparator. Checks if both responses are equal.

    """

    def compare(self, prediction, target, response_type, choices):
        """ Compares two response objects based on equality.
        When using the multiple-choice response type, the predictions and
        targets are interpreted as mask-vectors for the choices.
        In this case, the score corresponds to the overlap of the vectors.

        Parameters
        ----------
        prediction : tuple
            Response tuple A for comparison.

        target : tuple
            Response tuple B for comparison.
            
        response_type : string
            The response type of the prediction and target.
            
        choices : list(object)
            The choice options that were available for this comparison.

        Returns
        -------
        bool
            True if both objects are equal, false otherwise.

        """
        
        if response_type == "multiple-choice":

            string_choices = [tuple_to_string(x) for x in choices]
            string_preds = [tuple_to_string(x) for x in prediction]
            string_target = [tuple_to_string(x) for x in target]
            
            choices_pred = [x in string_preds for x in string_choices]
            choices_target = [x in string_target for x in string_choices]

            overlap = np.array(choices_pred) == np.array(choices_target)
            score = np.sum(overlap) / len(choices)
            return score
        
        return int(tuple_to_string(prediction) == tuple_to_string(target))

    def get_name(self):
        """ Returns the name of the comparator.

        Returns
        -------
        string
            Comparator name.

        """

        return 'Accuracy'



""" CCOBRA helper function.

"""

import copy

import numpy as np

def convert_to_basic_types(elem):
    """ Converts an element to primitive types. If the element
    is a list, the inner elements will be converted instead.
    The preference order is bool > int > float > string.

    Parameters
    ----------
    elem : Object
        Element to convert.

    Returns
    -------
    Object
        Primitive object representing the given element.

    """
    if isinstance(elem, int):
        return elem
    if isinstance(elem, float):
        return elem
    if isinstance(elem, bool):
        return elem
    if isinstance(elem, list):
        return [convert_to_basic_types(x) for x in elem]

    elem = str(elem)
    if elem == "True":
        return True
    if elem == "False":
        return False
    
    try:
        result = int(elem)
        return result
    except ValueError:
        try:
            result = float(elem)
            return result
        except ValueError:
            return elem
            
def tuple_to_string(tuptup):
    """ Converts a tuple to its string representation. Uses different separators (';', '/', '|') for
    different depths of the representation.

    Parameters
    ----------
    tuptup : list
        Tuple to convert to its string representation.

    Returns
    -------
    str
        String representation of the input tuple.

    """

    def join_deepest(tup, sep=';'):
        """ Recursive function to create the string representation for the deepest level of the
        tuptup list.

        Parameters
        ----------
        tup : object
            Element to join if list or list of lists.

        sep : str, optional
            Separation character to join the list elements by.

        Returns
        -------
        object
            List containing joined string in max depth. Str if input depth = 1.

        """

        if not isinstance(tup, list):
            return str(tup)
        if not isinstance(tup[0], list):
            return sep.join([str(x) for x in tup])

        for idx, val in enumerate(tup):
            tup[idx] = join_deepest(val, sep)
        return tup

    tup = copy.deepcopy(tuptup)
    tup = join_deepest(tup, ';')
    tup = join_deepest(tup, '/')
    tup = join_deepest(tup, '|')
    return tup


def unnest(tup):
    """ Unnests a nested tuple. If an element is insight nested lists, the function
    returns the element, otherwise the list is returned.

    Parameters
    ----------
    tup : list
        Nested tuples.

    Returns
    -------
    object
        Element or list without unneccessary nesting.

    """
    while True:
        if not isinstance(tup, list):
            return tup
        if len(tup) != 1:
            return tup
        
        tup = tup[0]


""" Propositional domain encoders.

"""

from ccobra import CCobraResponseEncoder

OPERATORS = [
    'not',
    'and',
    'or',
    'iff',
    'if'
]

class PropositionalResponseEncoder(CCobraResponseEncoder):
    """ Syllogistic encoder. Provides functions for abbreviating syllogistic responses.

    """

    @staticmethod
    def encode_clause(clause):
        """ Encodes a single clause by parsing the Polish normal form.

        Parameters
        ----------
        clause : list(str)
            Propositional clause in list representation (e.g., ['If', 'not', 'A', 'B']).

        Returns
        -------
        str
            String representation of the input clause (e.g., "~A -> B").

        """

        # Parse the clause in reverse
        stack = []
        for term in reversed(clause):
            norm_term = term.lower()
            if norm_term in OPERATORS:
                if norm_term == 'not':
                    stack.append('~{}'.format(stack.pop()))
                elif norm_term == 'and':
                    stack.append('({} & {})'.format(stack.pop(), stack.pop()))
                elif norm_term == 'or':
                    stack.append('({} | {})'.format(stack.pop(), stack.pop()))
                elif norm_term == 'iff':
                    stack.append('({} <=> {})'.format(stack.pop(), stack.pop()))
                elif norm_term == 'if':
                    stack.append('({} -> {})'.format(stack.pop(), stack.pop()))
            else:
                stack.append(term)

        assert len(stack) == 1, 'Error: {} -> {}'.format(clause, stack)
        return stack[0]

    @staticmethod
    def encode_response(response, task):
        """ Encodes a response to its propositional encoding.

        Parameters
        ----------
        response : list(str)
            Propositional response in list representation (e.g., ["not", "A"]).

        task : list(list(str))
            Propositional task as a list of clauses (e.g., [["If", "A", "B"], ["A"]]).

        Returns
        -------
        str
            String representation of the propositional response.

        """

        return PropositionalResponseEncoder.encode_clause(response[0])



""" Propositional submodule.

.. rubric:: Classes

.. autoclass:: PropositionalTaskEncoder
    :members:
.. autoclass:: PropositionalResponseEncoder
    :members:

"""

from .task_encoder_prop import PropositionalTaskEncoder
from .resp_encoder_prop import PropositionalResponseEncoder



""" CCOBRA model interface. Defines the general structure a model needs to
implement for being used in the CCOBRA framework.

"""

import logging


# Initialize module-level logger
logger = logging.getLogger(__name__)

class CCobraModel():
    """ Base class for CCOBRA models.

    """

    def __init__(self, name, supported_domains, supported_response_types):
        """ Base constructor of CCOBRA models.

        Parameters
        ----------
        name : str
            Unique name of the model. Will be used throughout the CCOBRA
            framework as a means for identifying the model.

        supported_domains : list(str)
            List containing the domains that are supported by the model (e.g.
            'syllogistic').

        supported_response_types : list(str)
            List containing the response types that are supported by the model
            (e.g., 'single-choice')

        """

        self.name = name
        self.supported_domains = supported_domains
        self.supported_response_types = supported_response_types

        self.evaluation_type = None

    def setup_environment(self, evaluation_type):
        """ Setup environment of the model using information about the prediction setting and the
        information provided during training.

        Parameters
        ----------
        evaluation_type : str
            Type of the CCOBRA evaluation (adaption, coverage).

        """

        self.evaluation_type = evaluation_type

    def start_participant(self, **kwargs):
        """ Callback to indicate participant start.

        """

        pass

    def end_participant(self, identifier, model_log, **kwargs):
        """ Hook for when participant simulation ends.

        Parameters
        ----------
        identifier : str or int
            Participant identifier.

        model_log : dict(str, object)
            Dictionary to allow the model to log information for the final output (e.g., parameter
            configurations for this participant).
        """

        pass

    def pre_train(self, dataset):
        """ Pre-trains the model based on given group data. This data is not necessarily other
        participants from the same experiment, but can also refer to an unrelated external dataset.
        The information supplied here represents the information known about the general population
        in the domain of interest.

        Parameters
        ----------
        dataset : list(list(dict(str, object)))
            Training data for the model. List of participants which each
            contain lists of tasks represented as dictionaries with the
            corresponding task information (e.g., the item container and
            given response).

        """

        pass

    def pre_train_person(self, dataset):
        """ Excerpt of the prediction data containing responses of the individual. Is supposed to
        combat the cold-start problem by supplying models with information about the exact
        individual to be predicted for. Allows to fit models to a specific individual. For
        example, in coverage settings, the responses given by the individual are supplied here.

        If not overriden by the model implemenation, uses adapt to perform the training.

        Parameters
        ----------
        dataset : list(dict(str, object))
            Training data for the model. List of tasks containing the items
            and corresponding responses.

        """

        if len(dataset) == 0:
            return

        for task_data in dataset:
            self.adapt(
                task_data['item'],
                task_data['response'],
                **dict([x for x in task_data.items() if x[0] not in ['item', 'response']])
            )

    def pre_person_background(self, dataset):
        """ Background information about the person to be predicted for. In contrast to the
        data supplied by pre_adapt, the data given here is not extracted from the test data.
        For example, could be responses given by the individual in question to an external
        independent experiment (e.g., participated in a spatial-relation experiment and later
        in a syllogistic experiment which serve as the test data).

        Parameters
        ----------
        dataset : list(dict(str, object))
            Training data for the model. List of tasks containing the items
            and corresponding responses.

        """

        pass

    def predict(self, item, **kwargs):
        """ Generates a single response prediction for a given task.

        Parameters
        ----------
        item : ccobra.Item
            Task information container. Holds the task text, response type,
            response choices, etc.

        Returns
        -------
        str
            Response prediction.

        """

        raise NotImplementedError()

    def adapt(self, item, target, **kwargs):
        """ Trains the model based on a given task-target combination.

        Parameters
        ----------
        item : ccobra.Item
            Task information container. Holds the task text, response type,
            response choices, etc.

        target : str
            True response given by the human reasoner.

        """

        pass



""" CCOBRA response comparator functionality.

.. rubric:: Classes

.. autoclass:: AbsDiffComparator
    :members:
.. autoclass:: EqualityComparator
    :members:
.. autoclass:: NVCComparator
    :members:
.. autoclass:: SquaredDiffComparator
    :members:

"""

from .absdiff import AbsDiffComparator
from .equality import EqualityComparator
from .nvc import NVCComparator
from .squareddiff import SquaredDiffComparator



""" CCOBRA Module

.. rubric:: Functions

.. autofunction:: convert_to_basic_types

.. autofunction:: tuple_to_string

.. autofunction:: unnest

.. rubric:: Submodules

.. autosummary::
   :toctree: _autosummary

   ccobra.benchmark
   ccobra.encoders
   ccobra.propositional
   ccobra.syllogistic
   ccobra.syllogistic_generalized

.. rubric:: Classes

.. autoclass:: CCobraComparator
    :members:
.. autoclass:: CCobraData
    :members:
.. autoclass:: CCobraModel
    :members:
.. autoclass:: CCobraResponseEncoder
    :members:
.. autoclass:: CCobraTaskEncoder
    :members:
.. autoclass:: Item
    :members:

"""


from .version import __version__

from .helper import convert_to_basic_types, tuple_to_string, unnest
from .data import CCobraData
from .item import Item
from .model import CCobraModel
from .encoder import CCobraTaskEncoder, CCobraResponseEncoder
from .comparator import CCobraComparator

from . import encoders
from . import benchmark
from . import propositional
from . import syllogistic
from . import syllogistic_generalized




""" Contains different comparator classes for model output data structures.

"""

class CCobraComparator():
    """ Comparator base class.

    """

    def compare(self, prediction, target, response_type, choices):
        """ Base comparison method.

        Parameters
        ----------
        prediction : object
            Prediction object for comparison.

        target : object
            Target object for comparison.
            
        response_type : string
            The response type of the prediction and target.
            
        choices : list(object)
            The choice options that were available for this comparison.

        Returns
        -------
        float
            Comparison result.

        """

        raise NotImplementedError()

    def get_name(self):
        """ Returns the name of the comparator.

        Returns
        -------
        string
            Comparator name.

        """

        raise NotImplementedError()



""" File containing classes required for benchmark handling.

"""

import json
import logging
import os

import pandas as pd

from . import comparators
from . import contextmanager
from . import modelimporter
from . import evaluation_handler
from .. import encoders
from .. import CCobraComparator, CCobraTaskEncoder, CCobraResponseEncoder
from ..data import CCobraData
from ..propositional.task_encoder_prop import PropositionalTaskEncoder
from ..syllogistic.task_encoder_syl import SyllogisticTaskEncoder

# Initialize logging
logger = logging.getLogger(__name__)

def fix_rel_path(path, base_path):
    """ Fixes relative paths by prepending the benchmark filepath.

    Parameters
    ----------
    path : str
        Path to fix.

    base_path : str
        Basepath used to fix relative paths with. Is prepended to the relative path.

    Returns
    -------
    str
        Fixed absolute path.

    """

    # Replace internal ccobra path
    if '%ccobra%' in path:
        package_path = os.path.split(os.path.split(__file__)[0])[0]
        path = path.replace('%ccobra%', package_path)

    if path and not os.path.isabs(path):
        return os.path.normpath(base_path + os.sep + path)

    return path

def fix_model_path(path, base_path=None):
    """ Fixes the model path by checking if the path directly refers to a python file. Otherwise
    searches for a subdirectory containing possible modules.

    Parameters
    ----------
    path : str
        Model path to fix.

    base_path : str, optional
        Base path to fix the model path with if it is relative.

    Returns
    -------
    str
        Path pointing to the file assumed to contain the model.

    """

    abs_path = path
    if base_path:
        abs_path = fix_rel_path(path, base_path)

    if os.path.isfile(abs_path) and abs_path[-2:] == "py":
        return abs_path

    python_files = []
    sub_directories = []
    for f_name in os.listdir(abs_path):
        f_path = os.path.join(abs_path, f_name)

        if os.path.isfile(f_path) and f_name[-2:] == "py":
            python_files.append(f_path)
        elif os.path.isdir(f_path) and f_name[0] != "." and f_name[:2] != "__":
            sub_directories.append(f_path)

    if python_files:
        return abs_path

    if len(sub_directories) == 1:
        python_files = [
            os.path.join(sub_directories[0], f) for f in os.listdir(
                sub_directories[0]) if os.path.isfile(
                    os.path.join(sub_directories[0], f)) and f[-2:] == "py"]
        if python_files:
            return sub_directories[0]

    raise ValueError("Could not identify model to load for '{}'".format(path))

def prepare_task_encoders(encoder_paths, base_path):
    """ Processes the task encoder information from the benchmark specification. Handles
    relative paths or path placeholders (e.g., '%ccobra%' mapping to the module directory of
    the local CCOBRA installation).

    Parameters
    ----------
    encoder_paths : dict(str, str)
        Dictionary mapping from domains to task encoders with absolute paths.

    Returns
    -------
    dict(str, str)
        Dictionary mapping from domains to paths containing task encoders.

    """

    encs = {}
    for domain, encoder_path in encoder_paths.items():
        # Normalize encoder path
        encoder_path = fix_rel_path(encoder_path, base_path)

        # To instantiate the encoder we need to change to its context (i.e., set the PATH variable
        # accordingly).
        enc = None
        with contextmanager.dir_context(encoder_path):
            imp = modelimporter.ModelImporter(encoder_path, superclass=CCobraTaskEncoder)
            enc = imp.instantiate()

        if not enc:
            raise ValueError('Failed to instantiate encoder class.')
        encs[domain] = enc

    return encs

def prepare_resp_encoders(encoder_paths, base_path):
    """ Processes the response encoder information from the benchmark specification. Handles
    relative paths or path placeholders (e.g., '%ccobra%' mapping to the module directory of
    the local CCOBRA installation).

    Parameters
    ----------
    encoder_paths : dict(str, str)
        Dictionary mapping from domains to response encoders with absolute paths.

    Returns
    -------
    dict(str, str)
        Dictionary mapping from domains to paths containing response encoders.

    """

    encs = {}
    for domain, encoder_path in encoder_paths.items():
        # Normalize encoder path
        encoder_path = fix_rel_path(encoder_path, base_path)

        # To instantiate the encoder we need to change to its context (i.e., set the PATH variable
        # accordingly).
        enc = None
        with contextmanager.dir_context(encoder_path):
            imp = modelimporter.ModelImporter(encoder_path, superclass=CCobraResponseEncoder)
            enc = imp.instantiate()

        if not enc:
            raise ValueError('Failed to instantiate encoder class.')
        encs[domain] = enc

    return encs

def prepare_comparator(comparator_path):
    """ Processes the comparator path from the benchmark specification. Imports the object
    dynamically.

    Parameters
    ----------
    comparator_path : str
        Path to the python script file containing the comparator definition.

    Returns
    -------
    ccobra.CCobraComparator
        Comparator object.

    """

    comp = None

    with contextmanager.dir_context(comparator_path):
        imp = modelimporter.ModelImporter(comparator_path, superclass=CCobraComparator)
        comp = imp.instantiate()

    if not comp:
        raise ValueError('Failed to instantiate comparator class.')

    return comp

class ModelInfo():
    """ Model information container. Contains the properties required to initialize and identify
    CCOBRA model instances.

    """

    def __init__(self, model_info, base_path, load_specific_class=None):
        """ Model initialization.

        Parameters
        ----------
        model_info : object
            Benchmark information about the model. Can either be string or dictionary.

        base_path : str
            Base path for handling relative path specifications.

        load_specific_class : str, optional
            Specific class name to load. Is used whenever multiple alternative CCOBRA model classes
            are specified within the model file.

        """

        #: Model filepath
        self.path = None

        #: String for overriding model name with
        self.override_name = None

        #: Class name for dynamic loading. Is used whenever multiple alternative CCOBRA model
        #: classes are specified within the model file.
        self.load_specific_class = load_specific_class

        #: Keyword arguments for the dynamic model instantiation
        self.args = {}

        if isinstance(model_info, str):
            self.path = fix_model_path(model_info, base_path)
        else:
            self.path = fix_model_path(model_info['filename'], base_path)
            self.override_name = model_info.get('override_name', self.override_name)
            self.args = model_info.get('args', self.args)
            self.load_specific_class = model_info.get('classname', self.load_specific_class)

    def __repr__(self):
        """ Generates a string representation for the model info container.

        Returns
        -------
        str
            ModelInfo string representation.

        """

        return str(self)

    def __str__(self):
        """ Generates a string representation for the model info container.

        Returns
        -------
        str
            ModelInfo string representation.

        """

        return 'path={}, override_name={}, load_specific_class={}, args={}'.format(
            self.path, self.override_name, self.load_specific_class, self.args)

class Benchmark():
    """ Benchmark class to handle and provide information from JSON benchmark specification files.

    """

    def __init__(self, json_path, argmodel=None, cached=False):
        """ Initializes the benchmark instance by reading the JSON benchmark specification file
        content.

        Parameters
        ----------
        json_path : str
            Path to the JSON benchmark specification file.

        argmodel : (str, str), optional
            Tuple containing the path to a specific model to load and the classname information.

        cached : bool, optional
            Flag to indicate whether the benchmark is cached or not. If true, the benchmark models
            are ignored.

        """

        logger.debug('Opening benchmark: "%s"', json_path)

        # Load raw benchmark file content
        self.json_content = None
        with open(json_path) as json_file:
            self.json_content = json.load(json_file)
        logger.debug('JSON content:\n%s', self.json_content)

        # Remove models in case of a cached run
        if cached:
            self.json_content['models'] = []
            logger.debug('Cached run. Removed models from benchmark')

        # Inject model added via arguments
        if argmodel != (None, None):
            argmodel_path = os.path.abspath(argmodel[0])
            self.json_content['models'].append(
                {"filename": argmodel_path, "classname": argmodel[1]})
            logger.debug('Injected model supplied via arguments')

        # Determine JSON path to fix relative path information
        self.base_path = os.path.dirname(os.path.abspath(json_path))
        logger.debug('base_path: %s', self.base_path)

        # Parse the JSON content
        self.parse_type()
        self.parse_auxiliary_evaluations()
        self.parse_models()
        self.parse_data()

        # Verify conditions for coverage
        if self.type == 'coverage':
            if self.data_pre_train_person is not None:
                raise ValueError('data.pre_train_person is not allowed in coverage evaluation.')

    def parse_type(self):
        """ Parses the benchmark type (prediction, adaption, coverage).

        """

        # Set type and validate
        self.type = self.json_content.get('type', 'adaption')
        if self.type not in ['prediction', 'adaption', 'coverage']:
            raise ValueError('Unsupported evaluation type: {}'.format(self.type))
        logger.debug('Evaluation type: %s', self.type)

    def parse_comparator(self, comparator_str):
        """ Parses the comparator information.

        Parameters
        ----------
        comparator_str : str
            Either is one of the library-defined comparator labels (equality, nvc, absdiff) or
            a path to a comparator implementation to load dynamically.

        Returns
        -------
        ccobra.CCobraComparator
            Comparator object.

        """

        # Create the comparator instance
        logger.debug('Comparator string: %s', comparator_str)
        if comparator_str == 'equality':
            return comparators.EqualityComparator()
        elif comparator_str == 'nvc':
            return comparators.NVCComparator()
        elif comparator_str == 'absdiff':
            return comparators.AbsDiffComparator()
        elif comparator_str == 'squareddiff':
            return comparators.SquaredDiffComparator()

        logger.debug('Nonlabel comparator string: %s', comparator_str)

        # Normalize path string
        comparator_str = fix_rel_path(comparator_str, self.base_path)
        logger.debug('Absolute comparator path: %s', comparator_str)

        if not os.path.isfile(comparator_str):
            raise ValueError('Comparator string is not a file: {}'.format(comparator_str))

        return prepare_comparator(comparator_str)

    def parse_auxiliary_evaluations(self):
        """ Parses auxiliary evaluation configurations from the benchmark content.

        """

        evaluations = self.json_content.get('aux_evaluations', [])

        task_encoders = self.json_content.get('task_encoders', {})
        if 'syllogistic' not in task_encoders:
            task_encoders['syllogistic'] = '%ccobra%/syllogistic/task_encoder_syl.py'
        if 'propositional' not in task_encoders:
            task_encoders['propositional'] = '%ccobra%/propositional/task_encoder_prop.py'

        resp_encoders = self.json_content.get('response_encoders', {})
        if 'syllogistic' not in resp_encoders:
            resp_encoders['syllogistic'] = '%ccobra%/syllogistic/resp_encoder_syl.py'
        if 'propositional' not in resp_encoders:
            resp_encoders['propositional'] = '%ccobra%/propositional/resp_encoder_prop.py'

        response_eval = {
            'data_column': 'response',
            'comparator': self.json_content.get('comparator', 'equality'),
            'prediction_fn_name': 'predict',
            'adapt_fn_name': 'adapt',
            'task_encoders': task_encoders,
            'response_encoders': resp_encoders
        }
        evaluations.insert(0, response_eval)

        evaluation_handlers = []
        evaluation_targets = []
        for eva in evaluations:
            task_encoders = None
            if 'task_encoders' in eva:
                task_encoders = prepare_task_encoders(eva['task_encoders'], self.base_path)

            resp_encoders = None
            if 'response_encoders' in eva:
                resp_encoders = prepare_resp_encoders(eva['response_encoders'], self.base_path)

            eh = evaluation_handler.EvaluationHandler(
                data_column=eva['data_column'],
                comparator=self.parse_comparator(eva.get('comparator', 'equality')),
                predict_fn_name=eva['prediction_fn_name'],
                adapt_fn_name=eva.get('adapt_fn_name', None),
                task_encoders=task_encoders,
                resp_encoders=resp_encoders
            )
            evaluation_handlers.append(eh)
            evaluation_targets.append(eva['data_column'])

            logger.debug('Added evaluation handler: %s', eh)

        self.evaluation_handlers = evaluation_handlers
        self.evaluation_targets = evaluation_targets

    def parse_data_path(self, path):
        """ Reads in a dataset CSV file and returns it as a pandas.DataFrame object. If a list
        of paths is supplied, the datasets are combined.

        Parameters
        ----------
        path : str
            Path to the data file.

        Returns
        -------
        (str, pandas.DataFrame)
            A tuple consisting of the filepath and the corresponding data frame. If a list of data
            paths was provided, the resulting string represents a ;-joined representation of the
            paths and the dataframe is the combination of the individual dataframes.

        """

        if not path:
            return None, None

        if isinstance(path, list):
            logger.debug('List data path encountered: %s', path)
            parts = [self.parse_data_path(x) for x in path]
            paths = ';'.join([x[0] for x in parts])

            # Combine the datasets
            comb_df = pd.concat([df for _, df in parts])
            return paths, comb_df

        logger.debug('Regular data path encountered: %s', path)

        # Resolve relative paths
        full_path = fix_rel_path(path, self.base_path)

        # Load the data and create CCOBRA container
        df = pd.read_csv(full_path)
        return full_path, df

    def parse_data(self):
        """ Parses the benchmark data information. Reads in an preprocesses the datasets.

        """

        # Verify information
        if 'data.test' not in self.json_content:
            raise ValueError('Test dataset (data.test) must be supplied.')

        # Parse training data fields
        self.data_pre_train_path, data_pre_train_df = self.parse_data_path(self.json_content.get('data.pre_train', ''))
        logger.debug('data_pre_train_path: %s', self.data_pre_train_path)
        self.data_pre_train_person_path, data_pre_train_person_df = self.parse_data_path(self.json_content.get('data.pre_train_person', ''))
        logger.debug('data_pre_train_person_path: %s', self.data_pre_train_person_path)
        self.data_pre_person_background_path, data_pre_person_background_df = self.parse_data_path(self.json_content.get('data.pre_person_background', ''))
        logger.debug('data_pre_person_background_path: %s', self.data_pre_person_background_path)

        # Parse test data field
        self.data_test_path, data_test_df = self.parse_data_path(self.json_content['data.test'])
        logger.debug('data_test_path: %s', self.data_test_path)

        # Filter person data so that only test ids are present
        test_ids = data_test_df['id'].unique()
        if data_pre_train_person_df is not None:
            data_pre_train_person_df = data_pre_train_person_df.loc[
                data_pre_train_person_df['id'].isin(test_ids)]
        if data_pre_person_background_df is not None:
            data_pre_person_background_df = data_pre_person_background_df.loc[
                data_pre_person_background_df['id'].isin(test_ids)]

        # Set corresponding data
        self.corresponding_data = self.json_content.get('corresponding_data', False)
        logger.debug('corresponding_data: %s', self.corresponding_data)

        # Construct CCOBRA datasets
        self.data_test = CCobraData(data_test_df, target_columns=self.evaluation_targets)
        self.data_pre_train = CCobraData(data_pre_train_df, target_columns=self.evaluation_targets) if data_pre_train_df is not None else None
        self.data_pre_train_person = CCobraData(data_pre_train_person_df, target_columns=self.evaluation_targets) if data_pre_train_person_df is not None else None
        self.data_pre_person_background = CCobraData(data_pre_person_background_df, target_columns=self.evaluation_targets) if data_pre_person_background_df is not None else None

        # In case of non-corresponding datasets, make sure that identifiers do not overlap by
        # offsetting the training data (ensures that test identifiers remain identifiable)
        if self.data_pre_train is not None and not self.corresponding_data:
            logger.debug('adjusting identifier offsets...')
            self.data_pre_train.prefix_identifiers()
        elif self.data_pre_train is not None and self.corresponding_data:
            logger.debug('extracting additional person data from comparing data_pre_train with data_test...')

            # Identify the columns which are present only in the training data
            merge = data_pre_train_df.merge(data_test_df, how='left', indicator=True)
            data_train_only_df = merge.loc[merge['_merge'] == 'left_only'].drop(columns=['_merge'])

            # Append domain related data to pre_train_person
            domain_related_df = data_train_only_df.loc[data_train_only_df['domain'].isin(self.data_test.domains)]

            if data_pre_train_person_df is not None:
                domain_related_df = pd.concat(domain_related_df, data_pre_train_person_df)

            if not domain_related_df.empty:
                self.data_pre_train_person = CCobraData(domain_related_df.drop(columns='_unique_id'), target_columns=self.evaluation_targets)

            # Append domain unrelated data to pre_person_background
            domain_unrelated_df = data_train_only_df.loc[~data_train_only_df['domain'].isin(self.data_test.domains)]

            if data_pre_person_background_df is not None:
                domain_unrelated_df = pd.concat(domain_unrelated_df, data_pre_person_background_df)

            if not domain_unrelated_df.empty:
                self.data_pre_person_background = CCobraData(domain_unrelated_df.drop(columns='_unique_id'), target_columns=self.evaluation_targets)

    def parse_models(self):
        """ Parses the benchmark model information.

        """

        # Prepare the models for loading
        self.models = [ModelInfo(x, self.base_path) for x in self.json_content['models']]
        logger.debug('models:\n%s', '\n'.join([str(x) for x in self.models]))

    def __str__(self):
        """ Generates a string representation of the benchmark information.

        """

        s = []
        s.append('Benchmark:')
        s.append('   type: {}'.format(self.type))
        s.append('   data paths:')
        s.append('      pre_train: {}'.format(self.data_pre_train_path))
        s.append('      pre_train_person: {}'.format(self.data_pre_train_person_path))
        s.append('      pre_person_background: {}'.format(self.data_pre_person_background_path))
        s.append('      test : {}'.format(self.data_test_path))
        s.append('   corresponding_data: {}'.format(self.corresponding_data))
        s.append('   models:')
        for idx, model in enumerate(self.models):
            s.append('      ({}) {}'.format(idx + 1, model))
        s.append('   evaluation handlers:')
        for eh in self.evaluation_handlers:
            s.append('      {}'.format(str(eh)))
        s.append('   evaluation targets: {}'.format(self.evaluation_targets))
        return '\n'.join(s)



""" CCOBRA evaluation module.

"""

import copy
import logging
import time

import numpy as np
import pandas as pd

from ..model import CCobraModel

from . import contextmanager
from . import modelimporter


# Initialize module-level logger
logger = logging.getLogger(__name__)

class Evaluator():
    """ CCOBRA evaluation routine.

    """

    def __init__(self, benchmark, is_silent=False, cache_df=None):
        """ Initializes the evaluator object by preparing the data representations and precomputing
        the required training and adaption steps.

        Parameters
        ----------
        benchmarks : ccobra.Benchmark
            Benchmark container.

        is_silent : bool, optional
            Flag indicating that output is supposed to be suppressed.

        cache_df : pandas.DataFrame, option
            Cache result dataframe.

        """

        logger.info('Setting up evaluator...')

        # Store the information
        self.benchmark = benchmark
        self.is_silent = is_silent
        self.cache_df = cache_df

        # Extract the dataset information
        self.dict_test = benchmark.data_test.to_eval_dict()

        self.dict_pre_train = None
        self.dict_pre_train_person = None
        self.dict_pre_person_background = None

        if benchmark.data_pre_train is not None:
            logger.debug('Supplied training data to evaluation.')
            self.dict_pre_train = benchmark.data_pre_train.to_eval_dict()
        if benchmark.data_pre_train_person is not None:
            logger.debug('Supplied person training data to evaluation.')
            self.dict_pre_train_person = benchmark.data_pre_train_person.to_eval_dict()
        if benchmark.data_pre_person_background is not None:
            logger.debug('Supplied person background data to evaluation.')
            self.dict_pre_person_background = benchmark.data_pre_person_background.to_eval_dict()

        if benchmark.type == 'coverage':
            self.dict_pre_train_person = self.dict_test

        # Extract the functionality to apply
        self.do_adapt = (benchmark.type == 'adaption')
        self.do_pre_train_global = (self.dict_pre_train is not None) and not benchmark.corresponding_data
        self.do_pre_train_leaveoneout = (self.dict_pre_train is not None) and benchmark.corresponding_data
        self.do_pre_train_person = (self.dict_pre_train_person is not None)
        self.do_pre_person_background = (self.dict_pre_person_background is not None)

        logger.debug('Evaluation ready:')
        logger.debug('   do_adapt: %s', self.do_adapt)
        logger.debug('   do_pre_train_global: %s', self.do_pre_train_global)
        logger.debug('   do_pre_train_leaveoneout: %s', self.do_pre_train_leaveoneout)
        logger.debug('   do_pre_train_person: %s', self.do_pre_train_person)
        logger.debug('   do_pre_person_background: %s', self.do_pre_person_background)

    def evaluate(self):
        """ Core evaluation routine.

        Returns
        -------
        pd.DataFrame
            Pandas dataframe containing the evaluation results.

        """

        logger.info('Starting evaluation routine...')

        model_logging_results = {}
        model_name_cache = set() if self.cache_df is None else set(self.cache_df['model'].unique())

        # Activate model context
        for model_idx, modelinfo in enumerate(self.benchmark.models):
            # Print the progress
            log_str = "Evaluating '{}' ({}/{})...".format(
                modelinfo.path, model_idx + 1, len(self.benchmark.models))
            logger.debug(''.join(['='] * 80))
            logger.info(log_str)
            logger.debug(''.join(['='] * 80))

            if not self.is_silent:
                print(log_str)

            # Initialize the dictionary for the models logging output
            model_logging_dict = {}

            # Setup model context
            with contextmanager.dir_context(modelinfo.path):
                # Dynamically import the CCOBRA model
                importer = modelimporter.ModelImporter(
                    modelinfo.path, CCobraModel,
                    load_specific_class=modelinfo.load_specific_class
                )

                # Instantiate and prepare the model for predictions
                pre_model = importer.instantiate(modelinfo.args)
                pre_model.setup_environment(self.benchmark.type)

                # Check if model is applicable to domains/response types
                self.check_model_applicability(pre_model)

                # Only use the model's name if no override is specified
                model_name = modelinfo.override_name
                if not model_name:
                    model_name = pre_model.name

                # Ensure that names are unique and show a warning if duplicates are detected
                original_model_name = model_name
                changed = False
                while model_name in model_name_cache:
                    model_name = model_name + '\''
                    changed = True
                model_name_cache.add(model_name)

                if changed:
                    logger.warning(
                        'Duplicate model name detected ("%s"). Changed to "%s".',
                        original_model_name, model_name
                    )

                # Only perform general pre-training if training data is
                # supplied and corresponding data is false. Otherwise, the
                # model has to be re-trained for each subject.
                if self.do_pre_train_global:
                    logger.debug('General pre-training for %s...', model_name)
                    pre_model.pre_train(list(self.dict_pre_train.values()))

                # Iterate subject
                for subj_key_identifier, subj_data in self.dict_test.items():
                    start_subject = time.time()

                    subj_id = subj_data[0]['item'].identifier
                    model = copy.deepcopy(pre_model)

                    # Set the model to new participant
                    model.start_participant(id=subj_id)

                    # Perform pre-training for individual subjects only if
                    # corresponding data is set to true
                    if self.do_pre_train_leaveoneout:
                        logger.debug('Individual pre-training for %s...', model_name)
                        cur_train_data = [
                            value for key, value in self.dict_pre_train.items() if key != subj_id]
                        model.pre_train(cur_train_data)

                    # Perform background fitting
                    if self.do_pre_person_background:
                        logger.debug('Person background training for %s...', model_name)
                        cur_train_data = self.dict_pre_person_background.get(subj_key_identifier, [])
                        model.pre_person_background(cur_train_data)

                    # Perform person training
                    if self.do_pre_train_person:
                        logger.debug('Person training for %s...', model_name)
                        subj_person_train_data = self.dict_pre_train_person.get(subj_key_identifier, [])
                        model.pre_train_person(subj_person_train_data)

                    # Iterate over individual tasks
                    start_eval = time.time()
                    for task_idx, task in enumerate(subj_data):
                        start_task = time.time()
                        logger.debug('Querying for task %s/%s...', task_idx + 1, len(subj_data))

                        # Integrity checks
                        assert task['item'].identifier == subj_id

                        # Query models for predictions
                        for eh in self.benchmark.evaluation_handlers:
                            target = task[eh.data_column]
                            eh.predict(model, model_name, task['item'], target, task['aux'])

                        # Perform model adaption
                        if self.do_adapt:
                            for eh in self.benchmark.evaluation_handlers:
                                target = task[eh.data_column]
                                eh.adapt(model, task['item'], task['full'])

                        logger.debug(
                            'Task {} took {:4f}s'.format(task_idx + 1, time.time() - start_task))

                    # Finalize subject evaluation and allow the model to store parameters
                    model_log = {}
                    model.end_participant(subj_id, model_log)
                    if len(model_log) > 0:
                        model_logging_dict[subj_id]= model_log

                    logger.debug('Subject evaluation took {:.4}s'.format(time.time() - start_eval))
                    logger.debug('Subject {} done. took {:.4}s'.format(
                        subj_id, time.time() - start_subject))

                # Save the models logging information if available
                if len(model_logging_dict) > 0:
                    model_logging_results[model_name] = model_logging_dict

                # Unload the imported model and its dependencies. Might cause garbage collection
                # issues
                importer.unimport()

        res_df = None
        on_list = [
            'model',
            'id',
            'domain',
            'response_type',
            'sequence',
            'task',
            'choices'
        ]

        for enc in self.benchmark.evaluation_handlers:
            if res_df is None:
                logger.debug('Preparing new result dataframe based on evaluation handler')
                res_df = enc.get_result_df()
            else:
                logger.debug('Adding evaluation handler result to result dataframe')
                res_df = res_df.merge(enc.get_result_df(), on=on_list, suffixes=('', '_' + enc.data_column))

        # Rename score column
        res_df = res_df.rename(columns={'score' : 'score_response'})

        # Integrate cache
        if self.cache_df is None:
            logger.debug('Empty cache. Returning only result dataframe.')
            return res_df, model_logging_results

        if res_df.empty:
            logger.debug('Empty result dataframe. Returning cache only.')
            return self.cache_df, {}

        logger.debug('Merging cache and result dataframe...')
        assert sorted(list(res_df)) == sorted(list(self.cache_df)), 'Incompatible cache'
        return pd.concat([res_df, self.cache_df]), model_logging_results

    def check_model_applicability(self, pre_model):
        """ Verifies the applicability of a model by checking its supported domains and response
        types and comparing them with the evaluation dataset.

        Parameters
        ----------
        pre_model : CCobraModel
            Model to check applicability for.

        Raises
        ------
        ValueError
            Exception thrown when model is not applicable to some domains or response types
            in the test data.

        """

        missing_domains = set(self.benchmark.data_test.domains) - set(pre_model.supported_domains)
        if missing_domains:
            raise ValueError(
                'Model {} is not applicable to domains {} found in ' \
                'the test dataset.'.format(
                    pre_model.name, missing_domains))

        missing_response_types = set(self.benchmark.data_test.response_types) - set(pre_model.supported_response_types)
        if missing_response_types:
            raise ValueError(
                'Model {} is not applicable to response_types {} ' \
                'found in the test dataset.'.format(
                    pre_model.name, missing_response_types))



""" Generalized syllogistic encoder functions.

"""

from ccobra import CCobraResponseEncoder


# Quantifier encodings
QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING = {
    'All': 'A',
    'No': 'E',
    'Few not': 'P',
    'Few': 'B',
    'Most': 'T',
    'Most not': 'D',
    'Many': 'K',
    'Many not': 'G',
    'Some': 'I',
    'Some not': 'O',
}

class GeneralizedSyllogisticResponseEncoder(CCobraResponseEncoder):
    """ Generalized Syllogistic encoder. Provides functions for abbreviating generalized
    syllogistic responses.

    """

    @staticmethod
    def encode_response(response, task):
        """ Encodes a response to its syllogistic encoding.

        Parameters
        ----------
        response : list(str)
            Syllogistc response in list representation (e.g., ['All', 'A', 'C'])

        task : list(list(str))
            Syllogistic task in list representation (e.g., [['All', 'A', 'B'], ['Some', 'B', 'C']]).

        Returns
        -------
        str
            Syllogistic response encoding (e.g., 'Aac').

        """

        if not isinstance(response[0], list):
            response = [response]

        if response[0] == 'NVC':
            return 'NVC'

        if response[0][0] == 'NVC':
            return 'NVC'

        object_sets = [set(x[1:]) for x in task]
        midterm = object_sets[0].intersection(object_sets[1])
        obj_a = object_sets[0] - midterm

        quant = QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING[response[0][0]]

        return quant + ('ac' if response[0][1] == list(obj_a)[0] else 'ca')



""" Squared differences comparator.

"""

import numpy as np

from ccobra import CCobraComparator, unnest


class SquaredDiffComparator(CCobraComparator):
    """ Squared differences comparator.

    """

    def compare(self, prediction, target, response_type, choices):
        """ Compares two response numbers based on their squared difference.

        Parameters
        ----------
        prediction : tuple
            Tuple containing number A.

        target : tuple
            Tuple containing number B.
            
        response_type : string
            The response type of the prediction and target.
            
        choices : list(object)
            The choice options that were available for this comparison.

        Returns
        -------
        float
            Squared difference between the differences.

        """

        if response_type == "multiple-choice":
            raise ValueError('The Squared Difference Comparator is incompatible with the multiple-choice response type.')

        inner_a = unnest(prediction)
        inner_b = unnest(target)
        
        if isinstance(inner_a, str):
            inner_a = float(inner_a)
        if isinstance(inner_b, str):
            inner_b = float(inner_b)
        if not isinstance(inner_a, (int, float)) or not isinstance(inner_b, (int, float)):
            raise ValueError('Incompatible value types for comparison.')

        return (inner_a - inner_b) ** 2

    def get_name(self):
        """ Returns the name of the comparator.

        Returns
        -------
        string
            Comparator name.

        """

        return "Squared Difference"



""" Generalized syllogistic encoder functions.

"""

from ccobra import CCobraTaskEncoder


# Quantifier encodings
QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING = {
    'All': 'A',
    'No': 'E',
    'Few not': 'P',
    'Few': 'B',
    'Most': 'T',
    'Most not': 'D',
    'Many': 'K',
    'Many not': 'G',
    'Some': 'I',
    'Some not': 'O',
}

class GeneralizedSyllogisticTaskEncoder(CCobraTaskEncoder):
    """ Generalized Syllogistic encoder. Provides functions for abbreviating generalized
    syllogistic tasks.

    """

    @staticmethod
    def encode_task(task):
        """ Encodes a task to its generalized syllogistic encoding.

        Parameters
        ----------
        task : list(list(str))
            List representation of the syllogism (e.g., [['All', 'A', 'B'], ['Some', 'B', 'C']]).

        Returns
        -------
        str
            Syllogistic task encoding (e.g., 'AI1').

        Raises
        ------
        ValueError
            If figure of syllogism cannot be determined.

        """

        prem_1, prem_2 = task

        quant1 = QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING[prem_1[0]]
        quant2 = QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING[prem_2[0]]
        figure = 1

        if prem_1[1] == prem_2[1]:
            figure = 4
        elif prem_1[2] == prem_2[1]:
            figure = 1
        elif prem_1[2] == prem_2[2]:
            figure = 3
        elif prem_1[1] == prem_2[2]:
            figure = 2
        else:
            raise ValueError('Could not determine figure of:', task)

        return quant1 + quant2 + str(figure)




""" Returns the response without further encoding. Useful, if values like reaction times
should be shown in the results.

"""

from ccobra import CCobraResponseEncoder

class IdentityResponseEncoder(CCobraResponseEncoder):
    """ Identity encoder. Returns the response without further encoding

    """

    @staticmethod
    def encode_response(response, task):
        """ Encodes a response to its syllogistic encoding.

        Parameters
        ----------
        response : list(str)
            Any response that should be returned

        task : list(list(str))
            Unused

        Returns
        -------
        str
            The response

        """
        return response



""" Syllogistic encoder functions.

"""

from ccobra import CCobraResponseEncoder


class SyllogisticResponseEncoder(CCobraResponseEncoder):
    """ Syllogistic encoder. Provides functions for abbreviating syllogistic responses.

    """

    @staticmethod
    def encode_response(response, task):
        """ Encodes a response to its syllogistic encoding.

        Parameters
        ----------
        response : list(str)
            Syllogistc response in list representation (e.g., ['All', 'A', 'C'])

        task : list(list(str))
            Syllogistic task in list representation (e.g., [['All', 'A', 'B'], ['Some', 'B', 'C']]).

        Returns
        -------
        str
            Syllogistic response encoding (e.g., 'Aac').

        """

        if not isinstance(response[0], list):
            response = [response]

        if response[0] == 'NVC':
            return 'NVC'

        if response[0][0] == 'NVC':
            return 'NVC'

        object_sets = [set(x[1:]) for x in task]
        midterm = object_sets[0].intersection(object_sets[1])
        obj_a = object_sets[0] - midterm

        quant = response[0][0].replace('All', 'A').replace(
            'Some not', 'O').replace('Some', 'I').replace('No', 'E')

        return quant + ('ac' if response[0][1] == list(obj_a)[0] else 'ca')



""" CCOBRA benchmark functionality.

.. rubric:: Submodules

.. autosummary::
   :toctree: _autosummary

   ccobra.benchmark.comparators

.. rubric:: Functions

.. autofunction:: dir_context
.. autofunction:: entry_point
.. autofunction:: fix_model_path
.. autofunction:: fix_rel_path
.. autofunction:: main
.. autofunction:: parse_arguments
.. autofunction:: silence_stdout

.. rubric:: Classes

.. autoclass:: Benchmark
    :members:
.. autoclass:: ModelInfo
    :members:
.. autoclass:: Evaluator
    :members:
.. autoclass:: ModelImporter
    :members:
.. autoclass:: EvaluationHandler
    :members:

"""

from . import comparators

from .benchmark import Benchmark, ModelInfo, fix_rel_path, fix_model_path
from .contextmanager import dir_context
from .evaluator import Evaluator
from .modelimporter import ModelImporter
from .runner import entry_point, parse_arguments, main, silence_stdout
from .evaluation_handler import EvaluationHandler



""" Imports a given python script and instantiates the contained model if
available.

Copyright 2018 Cognitive Computation Lab
University of Freiburg
Nicolas Riesterer <riestern@tf.uni-freiburg.de>
Daniel Brand <daniel.brand@cognition.uni-freiburg.de>

"""

import argparse
import codecs
import datetime
import logging
import os
import sys
import webbrowser
from contextlib import contextmanager

import pandas as pd

from . import benchmark as bmark
from . import evaluator
from .visualization import html_creator, viz_plot

from ..version import __version__


def parse_arguments():
    """ Parses the command line arguments for the benchmark runner.

    Returns
    -------
    dict
        Dictionary mapping from cmd arguments to values.

    """

    parser = argparse.ArgumentParser(description='CCOBRA version {}.'.format(__version__))
    parser.add_argument('benchmark', type=str, help='Benchmark file.')
    parser.add_argument('-m', '--model', type=str, help='Model file.')
    parser.add_argument(
        '-o', '--output', type=str, default='browser', help='Output style (browser/server).')
    parser.add_argument('-c', '--cache', type=str, help='Load specified cache file.')
    parser.add_argument('-s', '--save', type=str, help='Store results as csv table.')
    parser.add_argument(
        '-cn', '--classname', type=str, default=None,
        help='Load a specific class from a folder containing multiple classes.')
    parser.add_argument(
        '-ll', '--logginglevel', type=str, default='NONE',
        help='Set logging level [NONE, DEBUG, INFO, WARNING].'
    )

    args = vars(parser.parse_args())

    # Check for validity of command line arguments
    if not args['model'] and not args['benchmark']:
        print('ERROR: Must specify either model or benchmark.')
        parser.print_help()
        sys.exit(99)

    # Setup logging
    if args['logginglevel'].lower() == 'debug':
        logging.basicConfig(level=logging.DEBUG)
    elif args['logginglevel'].lower() == 'info':
        logging.basicConfig(level=logging.INFO)
    elif args['logginglevel'].lower() == 'warning':
        logging.basicConfig(level=logging.WARNING)

    return args

@contextmanager
def silence_stdout(silent, target=os.devnull):
    """ Contextmanager to silence stdout printing.

    Parameters
    ----------
    silent : bool
        Flag to indicate whether contextmanager should actually silence stdout.

    target : filepath, optional
        Target to redirect silenced stdout output to. Default is os.devnull.
        Can be modified to point to a log file instead.

    """

    new_target = open(target, 'w') if silent else sys.stdout
    old_target, sys.stdout = sys.stdout, new_target
    try:
        yield new_target
    finally:
        sys.stdout = old_target

def main(args):
    """ Main benchmark routine. Parses the arguments, loads models and data,
    runs the evaluation loop and produces the output.

    Parameters
    ----------
    args : dict
        Command line argument dictionary.

    """

    # Load cache information
    cache_df = None
    if args['cache']:
        cache_df = pd.read_csv(args['cache'])

    # Load the benchmark settings
    benchmark = bmark.Benchmark(
        args['benchmark'],
        argmodel=(args['model'], args['classname']),
        cached=(cache_df is not None)
    )

    # Run the model evaluation
    is_silent = (args['output'] in ['html', 'server'])
    eva = evaluator.Evaluator(benchmark, is_silent=is_silent, cache_df=cache_df)
    with silence_stdout(is_silent):
        res_df, model_log = eva.evaluate()

    if 'save' in args:
        res_df.to_csv(args['save'], index=False)

    # Create metrics dictionary
    #(TODO: check if there is a good way of dynamically specify the visualization)
    default_list = [
        viz_plot.AccuracyVisualizer(),
        viz_plot.BoxplotVisualizer(),
        viz_plot.SubjectTableVisualizer()
    ]
    metrics = []
    for idx, eva in enumerate(benchmark.evaluation_handlers):
        if idx == 0:
            metrics.append((
                eva, [
                    viz_plot.AccuracyVisualizer(),
                    viz_plot.BoxplotVisualizer(),
                    viz_plot.SubjectTableVisualizer(),
                    viz_plot.MFATableVisualizer(),
                    viz_plot.ModelLogVisualizer()
                ]))
        else:
            metrics.append((
                eva, default_list
            ))

    # Run the metric visualizer
    htmlcrtr = html_creator.HTMLCreator(metrics)

    # Prepare the benchmark output information and visualize the evaluation results
    path_pre_train = ''
    path_pre_train_person = ''
    path_pre_person_background = ''
    path_test = os.path.basename(benchmark.data_test_path)

    if benchmark.data_pre_train_path:
        path_pre_train = ';'.join([
            os.path.basename(x) for x in benchmark.data_pre_train_path.split(';')])
    if benchmark.data_pre_train_person_path:
        path_pre_train_person = ';'.join([
            os.path.basename(x) for x in benchmark.data_pre_train_person_path.split(';')])
    if benchmark.data_pre_person_background_path:
        path_pre_person_background = ';'.join([
            os.path.basename(x) for x in benchmark.data_pre_person_background_path.split(';')])

    benchmark_info = {
        'name': os.path.basename(args['benchmark']),
        'data.test': path_test,
        'data.pre_train': path_pre_train,
        'data.pre_train_person': path_pre_train_person,
        'data.pre_person_background': path_pre_person_background,
        'type': benchmark.type,
        'domains': list(res_df['domain'].unique()),
        'response_types': list(res_df['response_type'].unique()),
    }

    benchmark_info['corresponding_data'] = benchmark.corresponding_data

    # Generate the HTML output
    if args['output'] == 'server':
        html = htmlcrtr.to_html(res_df, benchmark_info, model_log, embedded=True)
        sys.stdout.buffer.write(html.encode('utf-8'))
    else:
        html = htmlcrtr.to_html(res_df, benchmark_info, model_log, embedded=False)

        # Save HTML output to file
        benchmark_filename = os.path.splitext(os.path.basename(args['benchmark']))[0]
        timestamp = datetime.datetime.now().strftime('%Y-%m-%d_%H-%M-%S')
        html_filename = '_'.join([benchmark_filename, timestamp, '.html'])
        html_filepath = os.path.join(benchmark.base_path, html_filename)

        with codecs.open(html_filepath, 'w', 'utf-8') as html_out:
            html_out.write(html)

        # Open HTML output in default browser
        webbrowser.open('file://' + os.path.realpath(html_filepath))

def entry_point():
    """ Entry point for the CCOBRA executables.

    """

    # Manually catch version command line argument
    if len(sys.argv) > 1 and sys.argv[1] == '--version':
        print('CCOBRA version {}'.format(__version__))
        exit()

    # Parse command line arguments
    args = parse_arguments()

    try:
        main(args)
    except Exception as exc:
        if args['output'] != 'html':
            raise
        msg = 'Error: ' + str(exc)
        if args['output'] == 'html':
            print('<p>{}</p><script>document.getElementById(\"result\").style.backgroundColor ' \
                '= \"Tomato\";</script>'.format(msg))
        else:
            print(exc)
        sys.exit()

if __name__ == '__main__':
    entry_point()



""" Module for the html output creation class.

"""

import os
import datetime
import json
import codecs

class HTMLCreator():
    """ Html output creator. Constructs the HTML string for displaying the CCOBRA evaluation
    results.

    """

    def __init__(self, evaluations):
        """ Initializes the html creator with a list of evaluations

        Parameters
        ----------
        evaluations : list(tuple(EvaluationHandler, list(PlotVisualizer)))
            List of tuples containing an EvaluationHandler object and the corresponding list of
            visualization objects, i.e., components for creating html snippets
            representing views (e.g., plots) on the data.
            
        """

        self.evaluations = evaluations

        # Load the template
        self.external_contents = {
            'template': '',
            'plotly': '',
            'html2canvas': '',
            'cssness': ''
        }

        ext_content_paths = {
            'template': 'template_page.html',
            'plotly': 'plotly-latest.min.js',
            'html2canvas': 'html2canvas.min.js',
            'cssness': 'template_page.css'
        }

        for key, path in ext_content_paths.items():
            path = os.path.dirname(__file__) + os.sep + path
            with codecs.open(path, "r", "utf-8") as file_handle:
                self.external_contents[key] = file_handle.read() + '\n'

    def to_html(self, result_df, benchmark, model_log, embedded=False):
        """ Generates the html output string.

        Parameters
        ----------
        result_df : pd.DataFrame
            DataFrame containing the CCOBRA evaluation results.

        benchmark : dict(str, object)
            Benchmark properties.
            
        model_log : dict(str, dict(str, object))
            Dictionary containing logging information that models supplied via end_participant.

        embedded : bool
            Flag indicating embedded usage. Removes CSS and window handling scripts from the
            resulting website.

        Returns
        -------
        str
            String containing html code representing the CCOBRA evaluation results.

        """

        result_data = json.dumps(result_df.to_csv(index=False).split('\n'))
        benchmark['date'] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M")

        # Construct the content for the website
        content = []
        css_dependencies = []
        
        for evaluation_tuple in self.evaluations:
            eval_handler, metrics = evaluation_tuple

            for metric in metrics:
                # Add dependencies
                if metric.template_CSS:
                    css_dependencies.append(metric.template_CSS)

                # Add HTML content div
                metric_html = metric.to_html(result_df, eval_handler, model_log)
                
                # Skip metrics if they have no information
                if metric_html is None:
                    continue
                
                metric_tab_data = (metric.shorttitle(eval_handler).lower().replace(' ', '-'), metric.shorttitle(eval_handler))

                metric_content = '<div id="{}-expand-bar" class="expand-bar">{}</div>'.format(
                    metric_tab_data[0], metric_tab_data[1])
                metric_content += '<div id="{}" class="expand-bar-content">{}</div>'.format(
                    metric_tab_data[0], metric_html)

                content.append(metric_content)

        # Generate auxiliary scripts
        scripts = []
        if not embedded:
            scripts.append('\n'.join([
                "           window.addEventListener('resize', function() {",
                "           var arr = document.getElementsByClassName('trigger_resize_script')",
                "           for (var n = 0; n < arr.length; n++)",
                "               eval(arr[n].innerHTML);",
                "           });"
            ]))

        # Construct CSS from visualizer dependencies
        css_content = ''
        if not embedded:
            css_content = self.external_contents['cssness']
            for fname in css_dependencies:
                path = os.path.dirname(__file__) + os.sep + fname
                with codecs.open(path, "r", "utf-8") as file_handle:
                    css_content += file_handle.read() + '\n'

        content_dict = {
            'CSSNESS': css_content,
            'PLOTLY_LIB': self.external_contents['plotly'],
            'HTML2CANVAS_LIB': self.external_contents['html2canvas'],
            'RESULT_DATA': result_data,
            'BENCHMARK': json.dumps(benchmark),
            'CONTENT': '\n\n'.join(content),
            'SCRIPTS': '\n\n'.join(scripts)
        }

        template = self.external_contents['template']
        for key, value in content_dict.items():
            template = template.replace('{{{{{}}}}}'.format(key), value)
        return template



""" Generalized Syllogistic submodule.

.. rubric:: Functions

.. autofunction:: decode_response
.. autofunction:: encode_response
.. autofunction:: encode_task

.. rubric:: Classes

.. autoclass:: GeneralizedSyllogism
    :members:
.. autoclass:: GeneralizedSyllogisticTaskEncoder
    :members:
.. autoclass:: GeneralizedSyllogisticResponseEncoder
    :members:

"""

from .task_encoder_sylgen import GeneralizedSyllogisticTaskEncoder, QUANTIFIERS_SYLLOGISTIC_GENERALIZED_ENCODING
from .resp_encoder_sylgen import GeneralizedSyllogisticResponseEncoder
from .syllogism_gen import decode_response, encode_response, encode_task, GeneralizedSyllogism



""" Syllogistic encoder functions.

"""

from ccobra import CCobraTaskEncoder

class SyllogisticTaskEncoder(CCobraTaskEncoder):
    """ Syllogistic encoder. Provides functions for abbreviating syllogistic tasks.

    """

    @staticmethod
    def encode_task(task):
        """ Encodes a task to its syllogistic encoding.

        Parameters
        ----------
        task : list(list(str))
            List representation of the syllogism (e.g., [['All', 'A', 'B'], ['Some', 'B', 'C']]).

        Returns
        -------
        str
            Syllogistic task encoding (e.g., 'AI1').

        Raises
        ------
        ValueError
            If figure of syllogism cannot be determined.

        """

        prem_1, prem_2 = task

        quant1 = prem_1[0].replace('All', 'A').replace(
            'Some not', 'O').replace('Some', 'I').replace('No', 'E')
        quant2 = prem_2[0].replace('All', 'A').replace(
            'Some not', 'O').replace('Some', 'I').replace('No', 'E')
        figure = 1

        if prem_1[1] == prem_2[1]:
            figure = 4
        elif prem_1[2] == prem_2[1]:
            figure = 1
        elif prem_1[2] == prem_2[2]:
            figure = 3
        elif prem_1[1] == prem_2[2]:
            figure = 2
        else:
            raise ValueError('Could not determine figure of:', task)

        return quant1 + quant2 + str(figure)

    @staticmethod
    def encode_response(response, task):
        """ Encodes a response to its syllogistic encoding.

        Parameters
        ----------
        response : list(str)
            Syllogistc response in list representation (e.g., ['All', 'A', 'C'])

        task : list(list(str))
            Syllogistic task in list representation (e.g., [['All', 'A', 'B'], ['Some', 'B', 'C']]).

        Returns
        -------
        str
            Syllogistic response encoding (e.g., 'Aac').

        """

        if not isinstance(response[0], list):
            response = [response]

        if response[0] == 'NVC':
            return 'NVC'

        if response[0][0] == 'NVC':
            return 'NVC'

        object_sets = [set(x[1:]) for x in task]
        midterm = object_sets[0].intersection(object_sets[1])
        obj_a = object_sets[0] - midterm

        quant = response[0][0].replace('All', 'A').replace(
            'Some not', 'O').replace('Some', 'I').replace('No', 'E')

        return quant + ('ac' if response[0][1] == list(obj_a)[0] else 'ca')



""" CCOBRA evaluation handler.

"""

import copy

import pandas as pd
import numpy as np

from .. import tuple_to_string

class EvaluationHandler():
    """ Evaluation handler class used to handle an evaluation setting.

    """
    def __init__(self, data_column, comparator, predict_fn_name, adapt_fn_name, task_encoders, resp_encoders):
        """ Initializes the Evaluation handler for a given data column and evaluation settings.

        Parameters
        ----------
        data_column : str
            Name of the data column to predict.

        comparator : ccobra.CCobraComparator
            Comparator to be used when comparing the prediction to the true value.

        predict_fn_name : str
            Name of the predict function within the models-

        adapt_fn_name : str
            Name of the adapt function within the models-

        task_encoders : dict(str, ccobra.CCobraTaskEncoder)
            Dictionary specifying the task encoders to be used for the domains in the dataset.

        resp_encoders : dict(str, ccobra.CCobraResponseEncoder)
            Dictionary specifying the response encoders to be used for the domains in the dataset.

        """
        self.data_column = data_column
        self.comparator = comparator
        self.predict_fn_name = predict_fn_name
        self.adapt_fn_name = adapt_fn_name
        self.task_encoders = task_encoders
        self.resp_encoders = resp_encoders

        # Prepare result dataframe
        self.result = []

    def predict(self, model, modelname, item, target, aux):
        """ Queries a given model for the prediction to a given task and manages the results.

        Parameters
        ----------
        model : ccobra.CCobraModel
            Model to query.

        modelname : str
            Name of the model in the results.

        item : ccobra.Item
            The item that the model should base the prediction on.

        target : tuple
            True response for the given item.

        aux : dict(str, object)
            Dictionary containing auxiliary information that should be passed to the model.

        """
        item = copy.deepcopy(item)
        aux = copy.deepcopy(aux)

        # Obtain the model prediction
        pred_fn = getattr(model, self.predict_fn_name, None)
        if pred_fn is None:
            raise NotImplementedError("{} has to be implemented in {}".format(self.predict_fn_name, modelname))

        prediction = pred_fn(item, **aux)
        
        score = self.comparator.compare(prediction, target, item.response_type, item.choices)

        # Collect the evaluation result data
        res_dict = {
            'model': modelname,
            'id': item.identifier,
            'domain': item.domain,
            'response_type': item.response_type,
            'sequence': item.sequence_number,
            'task': item.task_str,
            'choices': item.choices_str,
            'truth': tuple_to_string(target),
            'prediction': tuple_to_string(prediction),
            'score': score
        }

        if self.task_encoders:
            domain = res_dict['domain']
            res_dict['task_enc'] = self.task_encoders[domain].encode_task(item.task) if domain in self.task_encoders else np.nan

        if self.resp_encoders:
            domain = res_dict['domain']
            if item.response_type == "verify" or item.response_type == "accept":
                if len(item.choices) != 1:
                    raise ValueError("Only a single choice is allowed for response types 'verify' and 'accept'")
                
                truth_enc = np.nan
                prediction_enc = np.nan
                if domain in self.resp_encoders:
                    verification_target = item.choices[0]
                    verification_enc = self.resp_encoders[domain].encode_response(verification_target, item.task)
                
                    prediction_enc = "{};{}".format(verification_enc, prediction)
                    truth_enc = "{};{}".format(verification_enc, target)

                res_dict['truth_enc_{}'.format(self.data_column)] = truth_enc
                res_dict['prediction_enc_{}'.format(self.data_column)] = prediction_enc
            elif item.response_type == "multiple-choice":
                if not isinstance(prediction, list):
                    raise ValueError("A list of responses is required for multiple-choice predictions, but '{}' predicted '{}'".format(modelname, prediction))
                
                pred_encs = np.nan
                truth_encs = np.nan
                if domain in self.resp_encoders:
                    pred_encs = "|".join(sorted([self.resp_encoders[domain].encode_response(x, item.task) for x in prediction]))
                    truth_encs = "|".join(sorted([self.resp_encoders[domain].encode_response(x, item.task) for x in target]))
                res_dict['prediction_enc_{}'.format(self.data_column)] = pred_encs
                res_dict['truth_enc_{}'.format(self.data_column)] = truth_encs
                
            else:
                truth_enc = self.resp_encoders[domain].encode_response(target, item.task) if domain in self.resp_encoders else np.nan
                prediction_enc = self.resp_encoders[domain].encode_response(prediction, item.task) if domain in self.resp_encoders else np.nan
                res_dict['truth_enc_{}'.format(self.data_column)] = truth_enc
                res_dict['prediction_enc_{}'.format(self.data_column)] = prediction_enc

        self.result.append(res_dict)

    def adapt(self, model, item, full):
        """ Allows the given model to adapt to the true response to a given task.

        Parameters
        ----------
        model : ccobra.CCobraModel
            Model to query.

        item : ccobra.Item
            The item that the model should base the prediction on.

        full : dict(str, object)
            Dictionary containing the true response and the auxiliary information.

        """
        if self.adapt_fn_name is None:
            return

        item = copy.deepcopy(item)
        full = copy.deepcopy(full)

        target = full[self.data_column]
        aux = {x: y for x, y in full.items() if x != self.data_column}
        adapt_fn = getattr(model, self.adapt_fn_name, None)
        if adapt_fn is None:
            return

        adapt_fn(item, target, **aux)

    def get_result_df(self):
        """ Returns the results for the respective evaluation setting.

        Returns
        -------
        pd.DataFrame
            DataFrame containing the results for the evaluation setting.

        """
        return pd.DataFrame(self.result)

    def __repr__(self):
        s = 'EvaluationHandler(data_column={}, comparator={}, predict_fn_name={}, adapt_fn_name={}, task_encoders={}, resp_encoders={})'.format(
            self.data_column,
            self.comparator,
            self.predict_fn_name,
            self.adapt_fn_name,
            self.task_encoders,
            self.resp_encoders
        )
        return s

